(function() {
  var Extractor, Job, Mimeograph, Recognizer, Splitter, expose, fs, genkey, jobStatus, jobs, log, mimeograph, puts, redisfs, resque, spawn, _, _ref;
  var __hasProp = Object.prototype.hasOwnProperty, __extends = function(child, parent) {
    for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; }
    function ctor() { this.constructor = child; }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor;
    child.__super__ = parent.prototype;
    return child;
  }, __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice;
  mimeograph = exports;
  mimeograph.version = '1.0.0';
  fs = require('fs');
  resque = require('coffee-resque');
  spawn = require('child_process').spawn;
  redisfs = require('redisfs').redisfs;
  _ref = require('./utils'), _ = _ref._, log = _ref.log, puts = _ref.puts;
  expose = function(host, port) {
    redisfs = redisfs({
      namespace: 'mimeograph',
      prefix: 'mimeograph-',
      encoding: 'base64',
      host: host,
      port: port
    });
    global.redisfs = redisfs;
    global.redis = redisfs.redis;
    global.file2redis = _.bind(redisfs.file2redis, redisfs);
    return global.redis2file = _.bind(redisfs.redis2file, redisfs);
  };
  Job = (function() {
    function Job(context, callback) {
      var _ref2;
      this.context = context;
      this.callback = callback;
      _ref2 = this.context, this.key = _ref2.key, this.jobId = _ref2.jobId;
    }
    Job.prototype.complete = function(result) {
      return this.callback(_.extend(this.context, result));
    };
    Job.prototype.fail = function(err) {
      return this.callback(_.isString(err) ? new Error(err) : err);
    };
    return Job;
  })();
  Extractor = (function() {
    __extends(Extractor, Job);
    function Extractor() {
      Extractor.__super__.constructor.apply(this, arguments);
    }
    Extractor.prototype.perform = function() {
      log("extracting  - " + this.key);
      return redis2file(this.key, {
        file: this.key,
        deleteKey: false
      }, __bind(function(err) {
        if (err != null) {
          return this.fail(err);
        }
        return this.extract();
      }, this));
    };
    Extractor.prototype.extract = function() {
      var proc;
      proc = spawn('pdftotext', [this.key]);
      return proc.on('exit', __bind(function(code) {
        if (code !== 0) {
          return this.fail("pdftotext exit(" + code + ")");
        }
        return this.fetchText();
      }, this));
    };
    Extractor.prototype.fetchText = function() {
      this.file = "" + (_.basename(this.key)) + ".txt";
      return fs.readFile(this.file, 'utf8', __bind(function(err, text) {
        var file, _i, _len, _ref2;
        if (err != null) {
          return this.fail(err);
        }
        _ref2 = [this.file, this.key];
        for (_i = 0, _len = _ref2.length; _i < _len; _i++) {
          file = _ref2[_i];
          fs.unlink(file);
        }
        return this.complete({
          text: text
        });
      }, this));
    };
    return Extractor;
  })();
  Splitter = (function() {
    __extends(Splitter, Job);
    function Splitter(context, callback, pages) {
      this.context = context;
      this.callback = callback;
      this.pages = pages != null ? pages : [];
      Splitter.__super__.constructor.call(this, this.context, this.callback);
      this.basename = _.basename(this.key);
      this.args = ['-SDEVICE=pnggray', '-r600x600', '-sPAPERSIZE=letter', "-sOutputFile=" + this.basename + "-%04d.png", '-dTextAlphaBits=4', '-dBATCH', '-dNOPAUSE', '-dSAFER', '--', this.key];
    }
    Splitter.prototype.perform = function() {
      log("splitting   - " + this.key);
      return redis2file(this.key, {
        file: this.key
      }, __bind(function(err) {
        if (err != null) {
          return this.fail(err);
        }
        return this.split();
      }, this));
    };
    Splitter.prototype.split = function() {
      var proc;
      proc = spawn('gs', this.args);
      return proc.on('exit', __bind(function(code) {
        if (code !== 0) {
          return this.fail("gs exit(" + code + ")");
        }
        return this.fetchPages();
      }, this));
    };
    Splitter.prototype.fetchPages = function() {
      return fs.readdir('/tmp', __bind(function(err, files) {
        var file, _i, _len;
        if (err != null) {
          return this.fail(err);
        }
        for (_i = 0, _len = files.length; _i < _len; _i++) {
          file = files[_i];
          this.findPages(this.basename, "" + file);
        }
        fs.unlink(this.key);
        return this.complete({
          pages: this.pages
        });
      }, this));
    };
    Splitter.prototype.findPages = function(basename, file) {
      basename = basename.substr(basename.lastIndexOf('/') + 1);
      if (file.match("^" + basename + "-{1}")) {
        return this.pages.push("/tmp/" + file);
      }
    };
    return Splitter;
  })();
  Recognizer = (function() {
    __extends(Recognizer, Job);
    function Recognizer(context, callback) {
      this.context = context;
      this.callback = callback;
      Recognizer.__super__.constructor.call(this, this.context, this.callback);
      this.basename = _.basename(this.key);
      this.file = "" + this.basename + ".txt";
    }
    Recognizer.prototype.perform = function() {
      log("recognizing - " + this.key);
      return redis2file(this.key, {
        file: this.key
      }, __bind(function(err) {
        if (err != null) {
          return this.fail(err);
        }
        return this.recognize();
      }, this));
    };
    Recognizer.prototype.recognize = function() {
      return jobStatus(this.jobId, __bind(function(status) {
        var proc;
        if (status !== 'fail') {
          proc = spawn('tesseract', [this.key, this.basename]);
          return proc.on('exit', __bind(function(code) {
            if (code !== 0) {
              return this.fail("tesseract exit(" + code + ")");
            }
            return this.fetchText();
          }, this));
        } else {
          fs.unlink(this.key);
          return this.complete();
        }
      }, this));
    };
    Recognizer.prototype.fetchText = function() {
      return fs.readFile(this.file, 'utf8', __bind(function(err, text) {
        var file, _i, _len, _ref2;
        if (err != null) {
          return this.fail(err);
        }
        _ref2 = [this.key, this.file];
        for (_i = 0, _len = _ref2.length; _i < _len; _i++) {
          file = _ref2[_i];
          fs.unlink(file);
        }
        return this.complete({
          text: text,
          pageNumber: _.pageNumber(this.file)
        });
      }, this));
    };
    return Recognizer;
  })();
  jobs = {
    extract: function(context, callback) {
      return new Extractor(context, callback).perform();
    },
    split: function(context, callback) {
      return new Splitter(context, callback).perform();
    },
    ocr: function(context, callback) {
      return new Recognizer(context, callback).perform();
    }
  };
  genkey = function() {
    var args;
    args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    args.unshift("mimeograph:job");
    return args.join(':');
  };
  jobStatus = function(jobId, callback) {
    return redis.hget(genkey(jobId), 'status', __bind(function(err, status) {
      return callback(status);
    }, this));
  };
  Mimeograph = (function() {
    function Mimeograph(host, port, count, workers) {
      var i;
      this.workers = workers != null ? workers : [];
      this.end = __bind(this.end, this);
      this.error = __bind(this.error, this);
      this.success = __bind(this.success, this);
      expose(host, port);
      this.resque = resque.connect({
        redis: redis,
        namespace: 'resque:mimeograph'
      });
      for (i = 0; 0 <= count ? i < count : i > count; 0 <= count ? i++ : i--) {
        this.worker(i);
      }
      process.on('SIGINT', this.end);
      process.on('SIGTERM', this.end);
      process.on('SIGQUIT', this.end);
    }
    Mimeograph.prototype.start = function() {
      var worker, _i, _len, _ref2;
      _ref2 = this.workers;
      for (_i = 0, _len = _ref2.length; _i < _len; _i++) {
        worker = _ref2[_i];
        worker.start();
      }
      return log.warn("OK - mimeograph started with " + this.workers.length + " workers.");
    };
    Mimeograph.prototype.process = function(id, file) {
      var status;
      try {
        status = fs.lstatSync(file);
        if (!status.isFile()) {
          throw Error("'" + file + "' is not a file.");
        }
        if (id != null) {
          return this.createJob(id, file);
        } else {
          return redis.incr(genkey('ids'), __bind(function(err, id) {
            if (err != null) {
              return this.capture(err);
            }
            return this.createJob(_.lpad(id), file);
          }, this));
        }
      } catch (error) {
        puts.stderr("" + error.message);
        return this.end();
      }
    };
    Mimeograph.prototype.createJob = function(jobId, file) {
      var key;
      key = this.filename(jobId);
      redis.hset(genkey(jobId), 'started', _.now());
      return file2redis(file, {
        key: key
      }, __bind(function(err) {
        if (err != null) {
          return this.capture(err, {
            jobId: jobId
          });
        }
        this.enqueue('extract', key, jobId);
        puts.green("OK - created " + (genkey(jobId)) + " for file " + file);
        return this.end();
      }, this));
    };
    Mimeograph.prototype.filename = function(jobId) {
      return "/tmp/mimeograph-" + jobId + ".pdf";
    };
    Mimeograph.prototype.success = function(worker, queue, job, result) {
      switch (job["class"]) {
        case 'extract':
          return this.split(result);
        case 'split':
          return this.ocr(result);
        case 'ocr':
          return this.finish(result);
      }
    };
    Mimeograph.prototype.split = function(result) {
      var jobId, key, multi, text;
      jobId = result.jobId, key = result.key, text = result.text;
      if (!text.trim().length) {
        return this.enqueue('split', key, jobId);
      } else {
        multi = redis.multi();
        multi.del(key);
        multi.hset(genkey(jobId), 'text', text);
        return multi.exec(__bind(function(err, result) {
          if (err != null) {
            return this.capture(err, result);
          }
          return this.finalize(jobId);
        }, this));
      }
    };
    Mimeograph.prototype.ocr = function(result) {
      var jobId, pages;
      jobId = result.jobId, pages = result.pages;
      return redis.hset(genkey(jobId), 'num_pages', pages.length, __bind(function(err) {
        var page, _i, _len, _results;
        if (err != null) {
          return this.capture(err, result);
        }
        _results = [];
        for (_i = 0, _len = pages.length; _i < _len; _i++) {
          page = pages[_i];
          _results.push(this.storeAndEnqueue(page, 'ocr', jobId));
        }
        return _results;
      }, this));
    };
    Mimeograph.prototype.finish = function(result) {
      var jobId, multi, pageNumber, text;
      jobId = result.jobId, pageNumber = result.pageNumber, text = result.text;
      multi = redis.multi();
      multi.hset(genkey(jobId, 'text'), pageNumber, text);
      multi.hincrby(genkey(jobId), 'num_processed', 1);
      multi.hget(genkey(jobId), 'num_pages');
      return multi.exec(__bind(function(err, results) {
        var processed, total, _ref2;
        if (err != null) {
          return this.capture(err, result);
        }
        _ref2 = results.slice(1), processed = _ref2[0], total = _ref2[1];
        return this.checkComplete(jobId, processed, parseInt(total));
      }, this));
    };
    Mimeograph.prototype.checkComplete = function(jobId, processed, total) {
      if (processed === total) {
        return this.move2hash(jobId, __bind(function(err) {
          if (err != null) {
            return this.capture(err, result);
          }
          return this.finalize(jobId);
        }, this));
      }
    };
    Mimeograph.prototype.finalize = function(jobId) {
      var job, multi;
      job = genkey(jobId);
      multi = redis.multi();
      multi.hset(job, 'ended', _.now());
      multi.hset(job, 'status', 'complete');
      multi.hdel(job, 'num_processed');
      return multi.exec(__bind(function(err, results) {
        if (err != null) {
          return this.capture(err, result);
        }
        log.warn("finished    - " + job);
        return this.notify(job);
      }, this));
    };
    Mimeograph.prototype.notify = function(job, status) {
      if (status == null) {
        status = 'complete';
      }
      return redis.publish(job, "" + job + ":" + status, __bind(function(err, subscribers) {
        if (err != null) {
          return this.capture(err, result);
        }
        if (!subscribers) {
          return redis.sadd(genkey('completed'), job);
        }
      }, this));
    };
    Mimeograph.prototype.move2hash = function(jobId, callback, pages) {
      var errorskey, hashkey, multi;
      if (pages == null) {
        pages = [];
      }
      hashkey = genkey(jobId, 'text');
      errorskey = genkey(jobId, 'error_pages');
      multi = redis.multi();
      multi.zrange(errorskey, 0, -1);
      multi.hgetall(hashkey);
      return multi.exec(__bind(function(err, result) {
        var errors, field, hash, _i, _len, _ref2, _ref3;
        if (err != null) {
          return callback(err, {
            jobId: jobId
          });
        }
        _ref2 = result.slice(0), errors = _ref2[0], hash = _ref2[1];
        _ref3 = _.keys(hash).sort();
        for (_i = 0, _len = _ref3.length; _i < _len; _i++) {
          field = _ref3[_i];
          pages.push(hash[field]);
        }
        multi = redis.multi();
        multi.del(hashkey, errorskey);
        multi.hset(genkey(jobId), 'text', pages.join('\f'));
        if (!_.isEmpty(errors)) {
          multi.hset(genkey(jobId), 'error_pages', errors.join(','));
        }
        return multi.exec(__bind(function(err) {
          return callback(err);
        }, this));
      }, this));
    };
    Mimeograph.prototype.enqueue = function(job, key, jobId) {
      return this.resque.enqueue('mimeograph', job, [
        {
          key: key,
          jobId: jobId
        }
      ]);
    };
    Mimeograph.prototype.storeAndEnqueue = function() {
      var encoding, file, job, jobId, options, _i;
      file = arguments[0], encoding = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), job = arguments[_i++], jobId = arguments[_i++];
      options = {
        key: file
      };
      if (!_.isEmpty(encoding)) {
        options.encoding = encoding.shift();
      }
      return file2redis(file, options, __bind(function(err) {
        if (err != null) {
          return this.capture(err, {
            jobId: jobId,
            job: job
          });
        }
        return this.enqueue(job, file, jobId);
      }, this));
    };
    Mimeograph.prototype.error = function(error, worker, queue, job) {
      var jobId, multi, pageNumber;
      jobId = job.args[0].jobId;
      if (job["class"] === 'ocr') {
        pageNumber = _.pageNumber(job.args[0].key);
        log.err("error       - " + jobId + " " + job["class"] + " page: " + pageNumber);
        return redis.zadd(genkey(jobId, 'error_pages'), pageNumber, pageNumber, __bind(function(err) {
          if (err != null) {
            return this.capture(err);
          }
          return this.finish(_.extend(job, {
            jobId: jobId,
            pageNumber: pageNumber,
            text: ''
          }));
        }, this));
      } else {
        log.err("error       - " + jobId + " " + job["class"]);
        multi = redis.multi();
        multi.del(this.filename(jobId));
        multi.hset(genkey(jobId), 'status', 'failed');
        multi.hset(genkey(jobId), 'error', JSON.stringify(_.extend(job, error)));
        return multi.exec(__bind(function(err) {
          if (err != null) {
            return this.capture(err, {
              jobId: jobId
            });
          }
          return this.notify(genkey(jobId), 'fail');
        }, this));
      }
    };
    Mimeograph.prototype.worker = function(name) {
      var worker;
      this.workers.push(worker = this.resque.worker('mimeograph', jobs));
      worker.name = "mimeograph:" + name;
      worker.on('error', this.error);
      worker.on('success', this.success);
      return worker;
    };
    Mimeograph.prototype.end = function() {
      var worker, _i, _len, _ref2;
      log.warn('Shutting down!');
      _ref2 = this.workers;
      for (_i = 0, _len = _ref2.length; _i < _len; _i++) {
        worker = _ref2[_i];
        worker.end();
      }
      if (this.resque != null) {
        return this.resque.end();
      } else {
        return redisfs.end();
      }
    };
    Mimeograph.prototype.capture = function(err, meta) {
      return log.err("" + (JSON.stringify(meta)) + " " + (JSON.stringify(err)));
    };
    return Mimeograph;
  })();
  mimeograph.process = function(args, host, port) {
    var file, id;
    if (args.length === 2) {
      id = args.shift();
    }
    file = args.shift();
    return new Mimeograph(host, port).process(id, file);
  };
  mimeograph.start = function(host, port, workers) {
    if (workers == null) {
      workers = 5;
    }
    return new Mimeograph(host, port, workers).start();
  };
}).call(this);
